# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import functools
import pathlib

import click
from clapper.click import ResourceOption
from clapper.logging import setup

logger = setup(__name__.split(".")[0], format="%(levelname)s: %(message)s")


def reusable_options(f):
    """Wrap reusable predict script options (for ``experiment``).

    This decorator equips the target function ``f`` with all (reusable)
    ``predict`` script options.

    Parameters
    ----------
    f
        The target function to equip with options.  This function must have
        parameters that accept such options.

    Returns
    -------
        The decorated version of function ``f``
    """

    @click.option(
        "--output",
        "-o",
        help="""Path to a JSON file in which to save predictions for all samples in the
        input DataModule (leading directories are created if they do not
        exist).""",
        required=True,
        default="predictions.json",
        cls=ResourceOption,
        type=click.Path(
            file_okay=True,
            dir_okay=False,
            writable=True,
            path_type=pathlib.Path,
        ),
    )
    @click.option(
        "--model",
        "-m",
        help="""A lightning module instance implementing the network architecture
        (not the weights, necessarily) to be used for prediction.""",
        required=True,
        cls=ResourceOption,
    )
    @click.option(
        "--datamodule",
        "-d",
        help="""A lightning DataModule that will be asked for prediction data
        loaders. Typically, this includes all configured splits in a DataModule,
        however this is not a requirement.  A DataModule that returns a single
        dataloader for prediction (wrapped in a dictionary) is acceptable.""",
        required=True,
        cls=ResourceOption,
    )
    @click.option(
        "--batch-size",
        "-b",
        help="""Number of samples in every batch (this parameter affects memory
        requirements for the network).""",
        required=True,
        show_default=True,
        default=1,
        type=click.IntRange(min=1),
        cls=ResourceOption,
    )
    @click.option(
        "--device",
        "-d",
        help='A string indicating the device to use (e.g. "cpu" or "cuda:0")',
        show_default=True,
        required=True,
        default="cpu",
        cls=ResourceOption,
    )
    @click.option(
        "--weight",
        "-w",
        help="""Path or URL to pretrained model file (`.ckpt` extension),
        corresponding to the architecture set with `--model`.  Optionally, you may
        also pass a directory containing the result of a training session, in which
        case either the best (lowest validation) or latest model will be loaded.""",
        required=True,
        cls=ResourceOption,
        type=click.Path(
            exists=True,
            file_okay=True,
            dir_okay=True,
            readable=True,
            path_type=pathlib.Path,
        ),
    )
    @click.option(
        "--parallel",
        "-P",
        help="""Use multiprocessing for data loading: if set to -1 (default),
        disables multiprocessing data loading.  Set to 0 to enable as many data
        loading instances as processing cores available in the system.  Set to
        >= 1 to enable that many multiprocessing instances for data loading.""",
        type=click.IntRange(min=-1),
        show_default=True,
        required=True,
        default=-1,
        cls=ResourceOption,
    )
    @functools.wraps(f)
    def wrapper_reusable_options(*args, **kwargs):
        return f(*args, **kwargs)

    return wrapper_reusable_options


def predict(
    output,
    model,
    datamodule,
    batch_size,
    device,
    weight,
    parallel,
    **_,
) -> None:  # numpydoc ignore=PR01
    """Run inference (generates scores) on all input images, using a pre-trained model."""
    import json
    import shutil
    import typing

    from mednet.libs.classification.engine.predictor import run
    from mednet.libs.common.engine.device import DeviceManager
    from mednet.libs.common.utils.checkpointer import (
        get_checkpoint_to_run_inference,
    )

    from .utils import (
        device_properties,
        execution_metadata,
        model_summary,
        save_json_with_backup,
    )

    datamodule.set_chunk_size(batch_size, 1)
    datamodule.parallel = parallel
    datamodule.model_transforms = model.model_transforms

    datamodule.prepare_data()
    datamodule.setup(stage="predict")

    if weight.is_dir():
        weight = get_checkpoint_to_run_inference(weight)

    logger.info(f"Loading checkpoint from `{weight}`...")
    model = type(model).load_from_checkpoint(weight, strict=False)

    device_manager = DeviceManager(device)

    # register metadata
    json_data: dict[str, typing.Any] = execution_metadata()
    json_data.update(device_properties(device_manager.device_type))
    json_data.update(
        dict(
            database_name=datamodule.database_name,
            database_split=datamodule.split_name,
            model_name=model.name,
        ),
    )
    json_data.update(model_summary(model))
    json_data = {k.replace("_", "-"): v for k, v in json_data.items()}
    save_json_with_backup(output.with_suffix(".meta.json"), json_data)

    predictions = run(model, datamodule, device_manager)

    output.parent.mkdir(parents=True, exist_ok=True)
    if output.exists():
        backup = output.parent / (output.name + "~")
        logger.warning(
            f"Output predictions file `{str(output)}` exists - "
            f"backing it up to `{str(backup)}`...",
        )
        shutil.copy(output, backup)

    with output.open("w") as f:
        json.dump(predictions, f, indent=2)
    logger.info(f"Predictions saved to `{str(output)}`")
