# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import json
import pathlib
import typing

import numpy
import numpy.typing
import pytest
import torch
from mednet.libs.common.data.split import JSONDatabaseSplit
from mednet.libs.common.data.typing import DatabaseSplit


@pytest.fixture
def datadir(request) -> pathlib.Path:
    """Return the directory in which the test is sitting. Check the pytest
    documentation for more information.

    Parameters
    ----------
    request
        Information of the requesting test function.

    Returns
    -------
    pathlib.Path
        The directory in which the test is sitting.
    """

    return pathlib.Path(request.module.__file__).parents[0] / "data"


def pytest_configure(config):
    """This function is run once for pytest setup.

    Parameters
    ----------
    config
        Configuration values. Check the pytest documentation for more
        information.
    """

    config.addinivalue_line(
        "markers",
        "skip_if_rc_var_not_set(name): this mark skips the test if a certain "
        "~/.config/mednet.libs.classification.toml variable is not set",
    )

    config.addinivalue_line("markers", "slow: this mark indicates slow tests")


def pytest_runtest_setup(item):
    """This function is run for every test candidate in this directory.

    The test is run if this function returns ``None``.  To skip a test,
    call ``pytest.skip()``, specifying a reason.

    Parameters
    ----------
    item
        A test invocation item. Check the pytest documentation for more
        information.
    """

    from mednet.libs.classification.utils.rc import load_rc

    rc = load_rc()

    # iterates over all markers for the item being examined, get the first
    # argument and accumulate these names
    rc_names = [
        mark.args[0] for mark in item.iter_markers(name="skip_if_rc_var_not_set")
    ]

    # checks all names mentioned are set in ~/.config/mednet.libs.classification.toml, otherwise,
    # skip the test
    if rc_names:
        missing = [k for k in rc_names if rc.get(k) is None]
        if any(missing):
            pytest.skip(
                f"Test skipped because {', '.join(missing)} is **not** "
                f"set in ~/.config/mednet.libs.classification.toml",
            )


def rc_variable_set(name):
    from mednet.libs.classification.utils.rc import load_rc

    rc = load_rc()
    pytest.mark.skipif(
        name not in rc,
        reason=f"RC variable '{name}' is not set",
    )


@pytest.fixture(scope="session")
def temporary_basedir(tmp_path_factory):
    return tmp_path_factory.mktemp("test-cli")


class DatabaseCheckers:
    """Helpers for database tests."""

    @staticmethod
    def check_split(
        split: DatabaseSplit,
        lengths: dict[str, int],
        prefixes: typing.Sequence[str],
        possible_labels: typing.Sequence[int],
    ) -> None:
        """Run a simple consistency check on the data split.

        Parameters
        ----------
        split
            An instance of DatabaseSplit.
        lengths
            A dictionary that contains keys matching those of the split (this
            will be checked).  The values of the dictionary should correspond
            to the sizes of each of the datasets in the split.
        prefixes
            Each file named in a split should start with at least one of these
            prefixes.
        possible_labels
            These are the list of possible labels contained in any split.
        """

        assert len(split) == len(lengths)

        for k in lengths.keys():
            # dataset must have been declared
            assert k in split

            assert len(split[k]) == lengths[k]
            for s in split[k]:
                assert any([s[0].startswith(k) for k in prefixes]), (
                    f"Sample with name {s[0]} does not start with any of the "
                    f"prefixes in {prefixes}"
                )
                if isinstance(s[1], list):
                    assert all([k in possible_labels for k in s[1]])
                else:
                    assert s[1] in possible_labels

    @staticmethod
    def check_loaded_batch(
        batch,
        batch_size: int,
        color_planes: int,
        prefixes: typing.Sequence[str],
        possible_labels: typing.Sequence[int],
        expected_num_labels: int,
        expected_image_shape: tuple[int, ...] | None = None,
    ) -> None:
        """Check the consistency of an individual (loaded) batch.

        Parameters
        ----------
        batch
            The loaded batch to be checked.
        batch_size
            The mini-batch size.
        color_planes
            The number of color planes in the images.
        prefixes
            Each file named in a split should start with at least one of these
            prefixes.
        possible_labels
            These are the list of possible labels contained in any split.
        expected_num_labels
            The expected number of labels each sample should have.
        expected_image_shape
            The expected shape of the image (num_channels, width, height).
        """

        assert len(batch) == 2  # data, metadata

        assert isinstance(batch[0], torch.Tensor)
        assert batch[0].shape[0] == batch_size  # mini-batch size
        assert batch[0].shape[1] == color_planes

        if expected_image_shape:
            assert all(
                [data.shape == expected_image_shape for data in batch[0]],
            )

        assert isinstance(batch[1], dict)  # metadata
        assert len(batch[1]) == 2  # label and name

        assert "label" in batch[1]
        assert all([k in possible_labels for k in batch[1]["label"]])

        if expected_num_labels:
            assert len(batch[1]["label"]) == expected_num_labels

        assert "name" in batch[1]
        assert all(
            [any([k.startswith(j) for j in prefixes]) for k in batch[1]["name"]],
        )

        # use the code below to view generated images
        # from torchvision.transforms.functional import to_pil_image
        # to_pil_image(batch[0][0]).show()
        # __import__("pdb").set_trace()

    @staticmethod
    def _make_histo(data: numpy.typing.NDArray[numpy.uint8]) -> list[int]:
        from itertools import chain

        def _mk_single_channel(data: numpy.typing.NDArray[numpy.uint8]) -> list[int]:
            return numpy.histogram(data, bins=256, range=(0, 256))[0].tolist()

        return list(chain(*[_mk_single_channel(k) for k in data[0, :]]))

    @staticmethod
    def check_image_quality(
        datamodule,
        reference_histogram_file: pathlib.Path,
        compare_type: typing.Literal["equal", "statistical"] = "equal",
        pearson_coeff_threshold: float = 0.005,
    ) -> None:
        reference = {}
        for split, values in JSONDatabaseSplit(reference_histogram_file).items():
            reference[split] = {k[0]: k[1] for k in values}

        for split_name, loader in datamodule.predict_dataloader().items():
            for sample in loader:
                uint8_array = (255 * sample[0]).byte().numpy()
                histogram = DatabaseCheckers._make_histo(uint8_array)

                if sample[1]["name"][0] in reference[split_name]:
                    ref_histogram = reference[split_name].pop(sample[1]["name"][0])
                else:
                    continue

                if compare_type == "statistical":
                    # Compute pearson coefficients between histogram and
                    # reference and check the similarity within a certain
                    # threshold
                    pearson_coeffs = numpy.corrcoef(histogram, ref_histogram)
                    assert (1 - pearson_coeff_threshold) <= pearson_coeffs[0][1] <= 1

                else:
                    assert histogram == ref_histogram, (
                        f"Current and reference histograms for sample "
                        f"`{sample[1]['name'][0]}` at split `{split_name}` "
                        f"do not match: current = {histogram}, "
                        f"reference = {ref_histogram}"
                    )

        # all references must have been consumed
        for ref_split, left_values in reference.items():
            assert len(left_values) == 0, (
                f"Not all references at split `{ref_split}` were consumed: "
                f"{len(left_values)} are left"
            )

    @staticmethod
    def write_image_quality_histogram(
        datamodule,
        reference_histogram_file: pathlib.Path,
    ) -> None:
        data: dict[str, list] = {}
        for split_name, loader in datamodule.predict_dataloader().items():
            data[split_name] = []
            for sample in loader:
                uint8_array = (255 * sample[0]).byte().numpy()
                data[split_name].append(
                    [sample[1]["name"][0], DatabaseCheckers._make_histo(uint8_array)]
                )

        with reference_histogram_file.open("w") as f:
            json.dump(data, f)


@pytest.fixture
def database_checkers():
    return DatabaseCheckers
