# SPDX-FileCopyrightText: Copyright © 2024 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""COVD-DRIVE for Vessel Segmentation."""

import importlib.resources
import os
from pathlib import Path

import PIL.Image
from mednet.libs.common.data.datamodule import CachingDataModule
from mednet.libs.common.data.split import JSONDatabaseSplit
from mednet.libs.common.data.typing import DatabaseSplit, Sample
from mednet.libs.common.data.typing import RawDataLoader as _BaseRawDataLoader
from torchvision.transforms.functional import to_tensor

from ....utils.rc import load_rc

CONFIGURATION_KEY_DATADIR = "datadir." + (__name__.rsplit(".", 2)[-2])
"""Key to search for in the configuration file for the root directory of this
database."""


class RawDataLoader(_BaseRawDataLoader):
    """A specialized raw-data-loader for the Montgomery dataset."""

    datadir: str
    """This variable contains the base directory where the database raw data is
    stored."""

    def __init__(self):
        self.datadir = load_rc().get(
            CONFIGURATION_KEY_DATADIR, os.path.realpath(os.curdir)
        )

    def sample(self, sample: tuple[str, int]) -> Sample:
        """Load a single image sample from the disk.

        Parameters
        ----------
        sample
            A tuple containing the path suffix, within the dataset root folder,
            where to find the image to be loaded, and an integer, representing the
            sample label.

        Returns
        -------
            The sample representation.
        """

        image = PIL.Image.open(Path(self.datadir) / str(sample[0])).convert(
            mode="RGB"
        )
        tensor = to_tensor(image)
        label = PIL.Image.open(Path(self.datadir) / str(sample[1])).convert(
            mode="1", dither=None
        )
        mask = PIL.Image.open(Path(self.datadir) / str(sample[2])).convert(
            mode="1", dither=None
        )
        return tensor, dict(
            label=to_tensor(label), mask=to_tensor(mask), name=sample[0]
        )  # type: ignore[arg-type]

    def label(self, sample: tuple[str, int]) -> int:
        """Load a single image sample label from the disk.

        Parameters
        ----------
        sample
            A tuple containing the path suffix, within the dataset root folder,
            where to find the image to be loaded, and an integer, representing the
            sample label.

        Returns
        -------
        int
            The integer label associated with the sample.
        """
        return sample[1]


def make_split(basename: str) -> DatabaseSplit:
    """Return a database split for the Montgomery database.

    Parameters
    ----------
    basename
        Name of the .json file containing the split to load.

    Returns
    -------
        An instance of DatabaseSplit.
    """

    return JSONDatabaseSplit(
        importlib.resources.files(__name__.rsplit(".", 1)[0]).joinpath(basename)
    )


class DataModule(CachingDataModule):
    """DRIVE dataset for Vessel Segmentation.

    The DRIVE database has been established to enable comparative studies on
    segmentation of blood vessels in retinal images.

    * Reference: [DRIVE-2004]_
    * Original resolution (height x width): 584 x 565
    * Split reference: [DRIVE-2004]_
    * Protocol ``default``:

        * Training samples: 20 (including labels and masks)
        * Test samples: 20 (including labels from annotator 1 and masks)

    * Protocol ``second-annotator``:

        * Test samples: 20 (including labels from annotator 2 and masks)

    Parameters
    ----------
    split_filename
        Name of the .json file containing the split to load.
    """

    def __init__(self, split_filename: str):
        super().__init__(
            database_split=make_split(split_filename),
            raw_data_loader=RawDataLoader(),
        )
