# SPDX-FileCopyrightText: Copyright © 2024 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import click
from clapper.click import verbosity_option
from clapper.logging import setup
from clapper.rc import UserDefaults

logger = setup(__name__.split(".")[0], format="%(levelname)s: %(message)s")


def _echo(left: str, right: str, color: str = "white") -> None:
    s = [
        click.style(left, bold=True),
        click.style(": ", bold=True),
        click.style(right, fg=color),
    ]
    click.echo("".join(s))


def database_configuration(
    rc: UserDefaults, raw_databases: dict[str, dict[str, str]], lib_name: str
):
    """Display the database configuration for a particular library.

    Parameters
    ----------
    rc
        The user defaults read from the user .toml configuration file.
    raw_databases
        List of all supported (raw) databases.
    lib_name
        Name of the library to show configurations for.
    """

    if not rc.path.exists():
        _echo(f"{lib_name} configuration", f"{str(rc.path)} [MISSING]", "white")
    else:
        _echo(f"{lib_name} configuration", str(rc.path), "white")

    click.secho(f"configured {lib_name} databases:", bold=True)
    for k, v in raw_databases.items():
        if "datadir" not in v:
            # this database does not have a "datadir"
            continue

        if v["datadir"] is not None:
            _echo(f"  - {k} ({v['module']})", f"{v['datadir']}", "green")
        else:
            _echo(f"  - {k} ({v['module']})", "NOT installed", "red")


@click.command(
    epilog="""Examples:

1. Provide information about the current installation:

   .. code:: sh

      mednet info

""",
)
@verbosity_option(logger=logger, expose_value=False)
def info(
    **_,
) -> None:  # numpydoc ignore=PR01
    """Provide information about the current installation."""

    import typing

    from mednet.libs.classification.scripts.database import (
        _get_raw_databases as _get_raw_databases_classification,
    )
    from mednet.libs.classification.utils.rc import (
        load_rc as load_rc_classification,
    )
    from mednet.libs.common.scripts.utils import execution_metadata
    from mednet.libs.segmentation.scripts.database import (
        _get_raw_databases as _get_raw_databases_segmentation,
    )
    from mednet.libs.segmentation.utils.rc import (
        load_rc as load_rc_segmentation,
    )

    m = execution_metadata()

    _echo("mednet version", typing.cast(str, m["package-version"]), "green")
    _echo("platform", typing.cast(str, m["platform"]), "yellow")
    _echo(
        "accelerators",
        ", ".join(typing.cast(list[str], m["accelerators"])),
        "cyan",
    )

    database_configuration(
        load_rc_classification(),
        _get_raw_databases_classification(),
        "classification",
    )
    database_configuration(
        load_rc_segmentation(),
        _get_raw_databases_segmentation(),
        "segmentation",
    )

    click.secho("dependencies:", bold=True)
    python = typing.cast(dict[str, str], m["python"])
    _echo(
        "  - python",
        f"{python['version']} ({python['path']})",
        "white",
    )
    deps = typing.cast(dict[str, str], m["dependencies"])
    for name, version in deps.items():
        _echo(f"  - {name}", version, "white")
